package no.vegvesen.meldingsformidler.common.http

import java.net.Authenticator
import java.net.CookieHandler
import java.net.ProxySelector
import java.net.http.HttpClient
import java.net.http.HttpRequest
import java.net.http.HttpResponse
import java.time.Duration
import java.util.Optional
import java.util.concurrent.CompletableFuture
import java.util.concurrent.Executor
import javax.net.ssl.SSLContext
import javax.net.ssl.SSLParameters

class InterceptingHttpClient(private val delegate: HttpClient) : HttpClient() {

    private val defaultHeaders: MutableMap<String, () -> String?> = mutableMapOf()

    fun addHeaderProvider(headerName: String, valueProvider: () -> String?) {
        defaultHeaders[headerName] = valueProvider
    }

    override fun <T> send(request: HttpRequest, responseBodyHandler: HttpResponse.BodyHandler<T>): HttpResponse<T> {
        val modifiedRequest = injectHeaders(request)
        return delegate.send(modifiedRequest, responseBodyHandler)
    }

    override fun <T> sendAsync(request: HttpRequest, responseBodyHandler: HttpResponse.BodyHandler<T>): CompletableFuture<HttpResponse<T>> {
        val modifiedRequest = injectHeaders(request)
        return delegate.sendAsync(modifiedRequest, responseBodyHandler)
    }

    override fun <T : Any?> sendAsync(
        request: HttpRequest?,
        responseBodyHandler: HttpResponse.BodyHandler<T>?,
        pushPromiseHandler: HttpResponse.PushPromiseHandler<T>?
    ): CompletableFuture<HttpResponse<T>> {
        val modifiedRequest = injectHeaders(request!!)
        return delegate.sendAsync(modifiedRequest, responseBodyHandler, pushPromiseHandler)
    }
    private fun injectHeaders(request: HttpRequest): HttpRequest {
        val additionalHeaders = defaultHeaders.mapNotNull { (key, provider) -> provider()?.let { key to it } }

        return if (additionalHeaders.isEmpty()) {
            request
        } else {
            HttpRequest.newBuilder()
                .uri(request.uri())
                .method(request.method(), request.bodyPublisher().orElse(HttpRequest.BodyPublishers.noBody()))
                .headers(*mergeHeaders(request, additionalHeaders))
                .build()
        }
    }

    private fun mergeHeaders(request: HttpRequest, additionalHeaders: List<Pair<String, String>>): Array<String> {
        val headers = request.headers().map().flatMap { (key, values) -> values.map { key to it } }
        return (headers + additionalHeaders).flatMap { listOf(it.first, it.second) }.toTypedArray()
    }

    override fun followRedirects(): Redirect = delegate.followRedirects()

    override fun proxy(): Optional<ProxySelector> {
        return Optional.empty()
    }

    override fun version(): Version = delegate.version()
    override fun cookieHandler(): Optional<CookieHandler> = delegate.cookieHandler()
    override fun connectTimeout(): Optional<Duration> = delegate.connectTimeout()
    override fun sslContext(): SSLContext = delegate.sslContext()
    override fun sslParameters(): SSLParameters = delegate.sslParameters()
    override fun authenticator(): Optional<Authenticator> = delegate.authenticator()
    override fun executor(): Optional<Executor> = delegate.executor()

}
