package no.vegvesen.meldingsformidler.common.meldingsformidlerresult

import com.fasterxml.jackson.core.type.TypeReference
import com.fasterxml.jackson.databind.ObjectMapper
import org.slf4j.Logger
import java.net.http.HttpResponse

sealed interface MeldingsformidlerError<R>
data class UnexpectedError<R>(val throwable: Throwable, override val statusCode: Int) : MeldingsformidlerError<R>,
    HttpResult

data class Error<R>(val errorType: R) : MeldingsformidlerError<R>

@Suppress("unchecked_cast", "unused", "memberVisibilityCanBePrivate")
@JvmInline
value class MeldingsformidlerResult<T, R>(val value: Any?) {

    data class MeldingsformidlerFailure<R>(val error: MeldingsformidlerError<R>) {
        inline fun <reified T> fold(
            onError: (error: Error<R>) -> T, onUnexpectedError: (unexpectedError: UnexpectedError<R>) -> T
        ): T {
            return when (error) {
                is UnexpectedError -> onUnexpectedError(error)
                is Error -> onError(error)
            }
        }
    }

    fun isSuccess() = value !is MeldingsformidlerFailure<*>
    fun isFailure() = value is MeldingsformidlerFailure<*>

    fun getOrDefault(default: T): T = if (isSuccess()) value as T else default
    inline fun <T> getOrElse(func: (failure: MeldingsformidlerFailure<R>) -> T) =
        if (isSuccess()) value as T else func(value as MeldingsformidlerFailure<R>)

    fun getOrNull(): T? = if (isSuccess()) value as T else null
    fun failureOrNull(): MeldingsformidlerFailure<R>? = if (isFailure()) value as MeldingsformidlerFailure<R> else null

    fun <F, G> toDto(
        successToDto: (value: T) -> F = { value as F }, failureToDto: (error: R) -> G = { value as G }
    ): MeldingsformidlerResultDto<F, G> {
        return when (val error = failureOrNull()) {
            null -> MeldingsformidlerResultDto.success(successToDto(value as T))
            else -> when (error.error) {
                is UnexpectedError -> MeldingsformidlerResultDto.failure<T, R>(error.error.throwable) as MeldingsformidlerResultDto<F, G>
                is Error -> MeldingsformidlerResultDto.failure(failureToDto(error.error.errorType))
            }
        }
    }

    inline fun <F> onSuccess(block: (T) -> F): MeldingsformidlerResult<T, R> {
        if (isSuccess()) block(value as T)
        return this
    }

    inline fun <F> onFailure(block: (MeldingsformidlerFailure<R>) -> F): MeldingsformidlerResult<T, R> {
        if (isFailure()) block(value as MeldingsformidlerFailure<R>)
        return this
    }

    inline fun <F> onHttpError(block: (HttpResult.HttpStatus) -> F): MeldingsformidlerResult<T, R> {
        failureOrNull()?.fold(onError = { }, onUnexpectedError = { block(it.status()) })
        return this
    }

    inline fun <F> onException(block: (throwable: Throwable) -> F): MeldingsformidlerResult<T, R> {
        failureOrNull()?.fold(onError = {}, onUnexpectedError = { block(it.throwable) })
        return this
    }

    inline fun <F> onError(block: (error: R) -> F): MeldingsformidlerResult<T, R> {
        failureOrNull()?.fold(onError = { block(it.errorType) }, onUnexpectedError = { })
        return this
    }

    inline fun <reified F> fold(
        onSuccess: (success: T) -> F = { it as F }, onFailure: (failure: MeldingsformidlerFailure<R>) -> F
    ): F {
        return when (val error = failureOrNull()) {
            null -> onSuccess(value as T)
            else -> onFailure(error)
        }
    }


    companion object {
        fun <T, R> success(value: T): MeldingsformidlerResult<T, R> = MeldingsformidlerResult(value)
        fun <T, R> failure(error: MeldingsformidlerError<R>): MeldingsformidlerResult<T, R> =
            MeldingsformidlerResult(MeldingsformidlerFailure(error))

        fun <T, R> failure(error: R): MeldingsformidlerResult<T, R> =
            MeldingsformidlerResult(MeldingsformidlerFailure(Error(error)))

    }
}

@Suppress("unused")
inline fun <reified T, reified R, reified F, reified G> MeldingsformidlerResult<T, R>.toInternal(
    successToInternal: (T) -> F = { it as F },
    failureToInternal: (R) -> G = { it as G }
): MeldingsformidlerResult<F, G> {
    return when (val error = failureOrNull()) {
        null -> MeldingsformidlerResult.success(successToInternal(value as T))
        else -> when (error.error) {
            is UnexpectedError -> MeldingsformidlerResult.failure(UnexpectedError(error.error.throwable, error.error.statusCode))
            is Error -> MeldingsformidlerResult.failure(failureToInternal(error.error.errorType))
        }
    }
}

@Suppress("unused")
inline fun <reified T, reified R> MeldingsformidlerResult<T, R>.logOnError(
    log: Logger,
    prefix: String = "",
    httpErrorMessage: (HttpResult.HttpStatus) -> String = { "$prefix Error: Got http status ${it.name}" },
    exceptionMessage: (Throwable) -> String = { "$prefix Error: An unexpected error occurred" },
    errorMessage: (R) -> String = { "$prefix Error: $it" }
): MeldingsformidlerResult<T, R> {
    onHttpError { log.error(httpErrorMessage(it)) }
    onException { log.error(exceptionMessage(it), it) }
    onError { log.error(errorMessage(it)) }
    return this
}

@Suppress("unused")
inline fun <reified T, reified R> runCatchingMeldingsformidler(
    log: Logger? = null, block: () -> MeldingsformidlerResult<T, R>
): MeldingsformidlerResult<T, R> {
    return try {
        block()
    } catch (e: Throwable) {
        log?.error("An unexpected error occurred", e)
        MeldingsformidlerResult.failure(UnexpectedError(e, 500))
    }

}

@Suppress("unused")
inline fun <reified T, reified R> ObjectMapper.parseMeldingsformidlerResultDto(
    response: HttpResponse<String>,
): MeldingsformidlerResultDto<T, R> {
    return runCatching {
        this.readValue(
            response.body(),
            object : TypeReference<MeldingsformidlerResultDto<T, R>>() {})

    }.getOrElse {
        MeldingsformidlerResultDto.failure(it)
    }

}

