package no.vegvesen.meldingsformidler.common.meldingsformidlerresult

import com.fasterxml.jackson.annotation.JsonSubTypes
import com.fasterxml.jackson.annotation.JsonTypeInfo

@JsonTypeInfo(use = JsonTypeInfo.Id.NAME, include = JsonTypeInfo.As.PROPERTY, property = "result")
@JsonSubTypes(
    JsonSubTypes.Type(value = MeldingsformidlerResultSuccessDto::class, name = "success"),
    JsonSubTypes.Type(value = MeldingsformidlerResultFailureDto::class, name = "failure"),
    JsonSubTypes.Type(value = MeldingsformidlerResultExceptionDto::class, name = "exception")
)


sealed interface MeldingsformidlerResultDto<T, R> {
    companion object {
        fun <T, R> success(value: T): MeldingsformidlerResultDto<T, R> = MeldingsformidlerResultSuccessDto(value)
        fun <T, R> failure(error: R): MeldingsformidlerResultDto<T, R> = MeldingsformidlerResultFailureDto(error)
        fun <T, R> failure(exception: Throwable): MeldingsformidlerResultDto<T, R> =
            MeldingsformidlerResultExceptionDto(exception)
    }

}

inline fun <T, R, reified F, reified G> MeldingsformidlerResultDto<T, R>.toInternal(
    httpStatusCode: Int,
    success: (T) -> F = { it as F },
    failure: (R) -> G = { it as G },

    ): MeldingsformidlerResult<F, G> {
    if (httpStatusCode != 200) {
        return MeldingsformidlerResult.failure(
            UnexpectedError(
                Throwable("HTTP status code $httpStatusCode"),
                httpStatusCode
            )
        )
    }
    return when (this) {
        is MeldingsformidlerResultSuccessDto -> MeldingsformidlerResult.success(success(value))
        is MeldingsformidlerResultFailureDto -> MeldingsformidlerResult.failure(failure(error))
        is MeldingsformidlerResultExceptionDto -> MeldingsformidlerResult.failure(
            UnexpectedError(
                exception,
                httpStatusCode
            )
        )
    }
}

data class MeldingsformidlerResultSuccessDto<T, R>(
    val value: T,
) : MeldingsformidlerResultDto<T, R>

data class MeldingsformidlerResultFailureDto<T, R>(
    val error: R,
) : MeldingsformidlerResultDto<T, R>

data class MeldingsformidlerResultExceptionDto<T, R>(
    val exception: Throwable,
) : MeldingsformidlerResultDto<T, R>