package no.vegvesen.meldingsformidler.common.openamclient

import com.nimbusds.jwt.JWTClaimsSet
import com.nimbusds.jwt.SignedJWT
import java.io.IOException
import java.text.ParseException
import java.time.Clock
import java.time.Instant
import java.util.Date
import kotlin.time.Duration
import kotlin.time.Duration.Companion.ZERO
import kotlin.time.toKotlinDuration
import no.vegvesen.meldingsformidler.common.openamclient.TokenValidator.Result.EmptyToken
import no.vegvesen.meldingsformidler.common.openamclient.TokenValidator.Result.Expired
import no.vegvesen.meldingsformidler.common.openamclient.TokenValidator.Result.ShouldFetchNewToken
import no.vegvesen.meldingsformidler.common.openamclient.TokenValidator.Result.Valid
import org.slf4j.LoggerFactory

class OpenAmClient(
    private val oauth2Client: Oauth2Client,
    private val validator: TokenValidator = TokenValidator(),
) {
    private var bearerToken: BearerToken = BearerToken.emtpyToken()

    val log = LoggerFactory.getLogger(this::class.java)

    @Throws(OpenAmTokenException::class)
    fun token(): BearerToken {
        val validationResult = validator.validate(bearerToken)
        bearerToken = when (validationResult) {
            Valid -> bearerToken

            ShouldFetchNewToken -> tryGetNewToken()

            EmptyToken,
            Expired -> getNewTokenOrThrow()
        }

        return bearerToken
    }

    @Deprecated("Use token() instead, this will be removed in future versions, as this name is unclear")
    @Throws(OpenAmTokenException::class)
    fun get(): BearerToken {
        return token()
    }

    private fun tryGetNewToken(): BearerToken = try {
        val newToken = oauth2Client.getOrThrow()
        log.debug("OpenAM client: Token with Id:${newToken.id()} retrieved.")
        newToken
    } catch (_: Exception) {
        bearerToken
    }

    private fun getNewTokenOrThrow() = try {
        oauth2Client.getOrThrow()
    } catch (exception: Exception) {
        throw when (exception) {
            is IOException -> OpenAmTokenException("Could not connect to OpenAM", exception)
            is ParseException -> OpenAmTokenException("Could not parse token", exception)
            else -> OpenAmTokenException("Something went wrong when fetching OIDC access token", exception)
        }
    }

    data class BearerToken(val token: String, val clock: Clock = Clock.systemUTC()) {

        private val decodedToken: JWTClaimsSet? =
            if (token.isNotEmpty()) SignedJWT.parse(token).jwtClaimsSet else null

        override fun toString(): String {
            return "SVV Bearer token"
        }

        private fun expirationTime(): Date? = decodedToken?.expirationTime

        fun id() = decodedToken?.jwtid

        fun statensVegvesenRoles(): List<String?> {
            val svvRolesKey = "svvroles"
            return runCatching {
                decodedToken
                    ?.getStringListClaim(svvRolesKey)
            }.getOrElse {
                decodedToken
                    ?.getStringClaim(svvRolesKey)
                    ?.let { listOf(it) }
            }.orEmpty()
        }

        fun remainingValidTime(): Duration {
            return if (expirationTime() == null) {
                throw EmptyTokenexpirationTimeException()
            } else {
                val durationUntilExpired = java.time.Duration
                    .between(Instant.now(clock), expirationTime()?.toInstant())
                    .toKotlinDuration()

                if (durationUntilExpired.isNegative())
                    ZERO
                else
                    durationUntilExpired
            }
        }

        companion object {
            fun emtpyToken() = BearerToken("")
        }
    }
}

