package no.vegvesen.meldingsformidler.common.meldingsformidlerresult

import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.module.kotlin.registerKotlinModule
import no.vegvesen.meldingsformidler.common.meldingsformidlerresult.MeldingsformidlerResultTest.ResultType.Failure
import no.vegvesen.meldingsformidler.common.meldingsformidlerresult.MeldingsformidlerResultTest.ResultType.Success
import org.assertj.core.api.Assertions.assertThat
import org.junit.jupiter.api.Test
import java.net.URI
import java.net.http.HttpClient
import java.net.http.HttpHeaders
import java.net.http.HttpRequest
import java.net.http.HttpResponse
import java.util.Optional
import javax.net.ssl.SSLSession

class MeldingsformidlerResultTest {

    val objectMapper = ObjectMapper().registerKotlinModule()
    val exception = Throwable("An exception occurred")

    @Test
    fun `successful result maps correctly to success dto`() {
        val expected = MeldingsformidlerResultDto.success<SuccessTestDto, FailureTestDto>(SuccessTestDto("success"))
        val actual = runCatchingMeldingsformidler {
            createMeldingsformidlerResult(Success)
        }.toDto(
            SuccessTest::toDto,
            FailureTest::toDto
        )
        assertThat(expected).isEqualTo(actual)
    }

    @Test
    fun `failed result maps correctly to failure dto`() {
        val expected = MeldingsformidlerResultDto.failure<SuccessTestDto, FailureTestDto>(FailureTestDto("failure"))
        val actual = runCatchingMeldingsformidler {
            createMeldingsformidlerResult(Failure)
        }.toDto(
            SuccessTest::toDto,
            FailureTest::toDto
        )
        assertThat(expected).isEqualTo(actual)
    }

    @Test
    fun `successful dto maps correctly to success result`() {
        val expected = MeldingsformidlerResult.success<SuccessTest, FailureTest>(SuccessTest("success"))
        val actual = createMeldingsformidlerResultDto(Success)
            .toInternal(
                200,
                SuccessTestDto::toInternal,
                FailureTestDto::toInternal
            )
        assertThat(expected).isEqualTo(actual)
    }

    @Test
    fun `failure dto maps correctly to failure result`() {
        val expected = MeldingsformidlerResult.failure<SuccessTest, FailureTest>(FailureTest("failure"))
        val actual = createMeldingsformidlerResultDto(Failure)
            .toInternal(
                200,
                SuccessTestDto::toInternal,
                FailureTestDto::toInternal
            )
        assertThat(expected).isEqualTo(actual)
    }

    @Test
    fun `parsing successful dto correctly`() {
        val expected = MeldingsformidlerResultDto.success<SuccessTestDto, FailureTestDto>(SuccessTestDto("success"))
        val result = objectMapper.writeValueAsString(createMeldingsformidlerResultDto(Success))
        val response = MockHttpResponse(result)
        val actual = objectMapper.parseMeldingsformidlerResultDto<SuccessTestDto, FailureTestDto>(response)
        assertThat(actual).isEqualTo(expected)
    }

    @Test
    fun `parsing failure dto correctly`() {
        val expected = MeldingsformidlerResultDto.failure<SuccessTestDto, FailureTestDto>(FailureTestDto("failure"))
        val result = objectMapper.writeValueAsString(createMeldingsformidlerResultDto(Failure))
        val response = MockHttpResponse(result)
        val actual = objectMapper.parseMeldingsformidlerResultDto<SuccessTestDto, FailureTestDto>(response)
        assertThat(actual).isEqualTo(expected)
    }

    @Test
    fun `parsing successful dto and map to internal correctly`() {
        val expected = MeldingsformidlerResult.success<SuccessTest, FailureTest>(SuccessTest("success"))
        val actual = runCatchingMeldingsformidler {
            val result = objectMapper.writeValueAsString(createMeldingsformidlerResultDto(Success))
            val response = MockHttpResponse(result)
            objectMapper.parseMeldingsformidlerResultDto<SuccessTestDto, FailureTestDto>(response)
                .toInternal(
                    response.statusCode(),
                    SuccessTestDto::toInternal,
                    FailureTestDto::toInternal
                )
        }
        assertThat(actual).isEqualTo(expected)
    }

    @Test
    fun `parsing response with http status 500 maps to failure`() {
        val expected = MeldingsformidlerResult.failure<SuccessTest, FailureTest>(
            UnexpectedError(
                Throwable("HTTP status code 500"),
                500
            )
        )
        val actual = runCatchingMeldingsformidler {
            val body = objectMapper.writeValueAsString(createMeldingsformidlerResultDto(Success))
            val response = MockHttpResponse(
                responseBody = body,
                statusCode = 500,
            )
            objectMapper.parseMeldingsformidlerResultDto<SuccessTestDto, FailureTestDto>(response)
                .toInternal(
                    response.statusCode(),
                    SuccessTestDto::toInternal,
                    FailureTestDto::toInternal
                )
        }

        assertThat(actual.toString()).isEqualTo(expected.toString())
    }

    @Test
    fun `parsing response with exception maps to failure`() {
        val actual = runCatchingMeldingsformidler {
            val body = objectMapper.writeValueAsString("dette er ikke gyldig json")
            val response = MockHttpResponse(
                responseBody = body,
                statusCode = 200,
            )
            objectMapper.parseMeldingsformidlerResultDto<SuccessTestDto, FailureTestDto>(response)
                .toInternal(
                    response.statusCode(),
                    SuccessTestDto::toInternal,
                    FailureTestDto::toInternal
                )
        }
        assertThat(actual.isFailure()).isTrue()
    }

    @Test
    fun `parsing exception response correctly`() {
        val expected = MeldingsformidlerResult.failure<SuccessTest, FailureTest>(
            UnexpectedError(
                exception,
                200
            )
        )
        val actual = runCatchingMeldingsformidler {
            val body = objectMapper.writeValueAsString(createMeldingsformidlerResultDto(ResultType.Exception))
            val response = MockHttpResponse(
                responseBody = body,
                statusCode = 200,
            )
            objectMapper.parseMeldingsformidlerResultDto<SuccessTestDto, FailureTestDto>(response)
                .toInternal(
                    response.statusCode(),
                    SuccessTestDto::toInternal,
                    FailureTestDto::toInternal
                )
        }
        assertThat(actual.toString()).isEqualTo(expected.toString())
    }

    private class MockHttpResponse(
        private val responseBody: String,
        private val statusCode: Int = 200,
        private val headers: Map<String, List<String>> = emptyMap()
    ) : HttpResponse<String> {
        override fun statusCode(): Int = statusCode
        override fun request(): HttpRequest {
            TODO("Not yet implemented")
        }

        override fun previousResponse(): Optional<HttpResponse<String>> {
            TODO("Not yet implemented")
        }

        override fun headers(): HttpHeaders {
            return HttpHeaders.of(headers) { key, value -> true }
        }

        override fun body(): String = responseBody
        override fun sslSession(): Optional<SSLSession> {
            TODO("Not yet implemented")
        }

        override fun uri(): URI {
            TODO("Not yet implemented")
        }

        override fun version(): HttpClient.Version {
            TODO("Not yet implemented")
        }

    }

    private enum class ResultType {
        Success, Failure, Exception
    }

    private fun createMeldingsformidlerResultDto(resultType: ResultType): MeldingsformidlerResultDto<SuccessTestDto, FailureTestDto> {
        return when (resultType) {
            Success -> MeldingsformidlerResultDto.success(SuccessTestDto("success"))
            Failure -> MeldingsformidlerResultDto.failure(FailureTestDto("failure"))
            ResultType.Exception -> MeldingsformidlerResultDto.failure(exception)
        }
    }

    private fun createMeldingsformidlerResult(resultType: ResultType): MeldingsformidlerResult<SuccessTest, FailureTest> {
        return when (resultType) {
            Success -> MeldingsformidlerResult.success(SuccessTest("success"))
            Failure -> MeldingsformidlerResult.failure(FailureTest("failure"))
            ResultType.Exception -> MeldingsformidlerResult.failure(UnexpectedError(exception, 500))
        }

    }


    private data class SuccessTest(val value: String) {
        fun toDto(): SuccessTestDto {
            return SuccessTestDto(value)
        }
    }

    private data class SuccessTestDto(val value: String) {
        fun toInternal(): SuccessTest {
            return SuccessTest(value)
        }
    }

    private data class FailureTest(val value: String) {
        fun toDto(): FailureTestDto {
            return FailureTestDto(value)
        }
    }

    private data class FailureTestDto(val value: String) {
        fun toInternal(): FailureTest {
            return FailureTest(value)
        }
    }
}